require "Automaten"
require "Grammatiken"
require "Warshall.rb"
#
# Author:: Bauer, Sebastian (mailto:sebastian.bauer@unibw.de)
# Author:: Jede, Alexander (mailto:alexander.jede@unibw.de)
# Author:: Wickmann, Ren (mailto:rene.wickmann@unibw.de)
# Copyright:: GPLv3

##
# Diese Klasse reprsentiert die bergangstabelle eines (S)LR-Automaten.
# Sie arbeitet mit Automaten im DEA-Format.
# Jedem Zustand wird ein reprsentativer Name der Form K[i] mit i  |N zugeordnet.
# Wenn kein bergang zu einem anderem Zustand vorhanden ist, wird dies durch "-" dargestellt.
# 
class UTable
  ##
  # Der Konstruktor erwartet einen (S)LR-Automaten im DEA-Format.
  # Die bergangstabelle wird intern in einem zweidimensionalem Hash berechnet
  def initialize(dea)
    # Alle Zustnde in einem Hash speichern und ihnen einen "ordentlichen" Namen
    # nach folgendem Muster zuweissen: K[i] wobei i eine Nummer ist.
    # Eine Tablle ist ein Hash, mit Namen von Zustnden (das ist der Key), denen die bergnge
    # zugewiesen werden. Ein bergang ist ein Key (das Symbol) dem der neue Zustand (Name) als Wert
    # zugewiesen wird
    i = 1;
    @namehs = Hash.new
    @tabelle = Hash.new
    @zusths = Hash.new
    dea.zustaende_each{ |zust|
      @namehs["K#{i}"] = zust.to_s
      @zusths["K#{i}"] = zust
      @tabelle["K#{i}"] = Hash.new("-")
      i +=1
    }
    @uSymbole = Hash.new
		a = Symb.new("#")
			@uSymbole[a] = a.to_s
	@uSymbole.merge!(dea.getGrammar.getNonterminals.invert)
	@uSymbole.merge!(dea.getGrammar.getTerminals.invert)
    dea.uebergaenge_each{|konf|
      # Jede Konfiguration wird aufgespalten und die Information wird wie oben  beschrieben abgelegt
      @tabelle[@namehs.invert[konf[0].to_s]][konf[1].to_s] = @namehs.invert[konf[2].to_s]
      #@uSymbole[konf[1]] = konf[1].to_s
    }
		 
  end
  
  ##
  # Liefert den Hash mit den Namen der Zustnde.
  # Der Schlssel ist der Name eines Zustandes der Form K[i], wobei i eine natrliche Zahl ist.
  # Diesem Schlssel wird ein Zustand als String (Zustand.to_s) zugeordnet.
  # Bsp.: [ "K1" => "{ [S -> a .+ b],[S -> a .* b] }" ]
  def getnamehs
	@namehs
  end
  
  ##
  # Liefert die bergangstabelle des Automaten.
  # Bei der Tabelle handelt es sich um einen Hash, der als Schlssen einen 
  # Zustandsnamen hat (sie dazu getnamehs) und der dazugehrige Wert ist ebenfalls 
  # ein Hash. Dieser Hash stellt eine Zeile der Tabelle dar.
  # Der Schlssel ist hier der String eines Symbols, dem ein Name eines Zustand als Wert zugeordnet wird.
  # Beispiel fr solch eine Zeile wre [ "a" => "K2", "b" => "K1"].
  # Sollte fr ein Symbol kein bergang vorhanden sein, wird ein "-" zurckgegeben.
  #
  def gettabelle
	@tabelle
  end
	
  ##
  # Liefert einen Hash, in dem jedem Symbol, dem Schlssel, 
  # der zugehrige String zugeordnet wird. 
  # Bsp.: a = Symb.new("A")
  #       [ a => "A" ]
  def getsymbole
	@uSymbole
  end
	
  ##
  # Liefert die bergangstabelle als formatierten String.
  # Die Ausgangszustnde stehen Links in der Tabelle als Zeilen.
  # Die bergangssymbole stehen Oben als Spalten, die Zielzustnde
  # sind die Eintrge in dieser Matrix. Sollte kein Zielzustand vorhanden sein, wird dies durch 
  # ein "-" dargestellt.
  def to_s
    symbole = @uSymbole.keys
    maxZlaenge = (@namehs.size).to_s.length+1
    spaltenLaenge = Array.new
    # die Breite fr jede Spalte berechnen
    # richtet sich nach dem lngstem Eintrag in einer Spalte
    symbole.each{|symb|
      maxLaenge = symb.to_s.length
      @tabelle.each_value{|zeile|
        (zeile[symb.to_s].length > maxLaenge) ? (maxLaenge = zeile[symb.to_s].length) : () 
      }
      spaltenLaenge.push(maxLaenge)
    }
    
    # Kopfzeile
    erg = " " *(maxZlaenge+1)+"| "
    spaltenLaenge.each_index{|index|
      erg += (symbole[index].to_s) + " " *( spaltenLaenge[index] - symbole[index].to_s.length + 1)
    }
    erg += "\n"
    # Trennzeile
    erg += "-" *(erg.length-1) + "\n"
    #alle restlichen Zeilen ausgeben/hinzufgen
    @namehs.each_key{|zName|
      erg += zName + (" " *(maxZlaenge - zName.length + 1)) + "| "
      symbole.each_index{|index|
      folgeZust = (@tabelle[zName])[symbole[index].to_s]
      erg += folgeZust + " " *(spaltenLaenge[index] - folgeZust.length + 1)
      }
      erg +="\n"
    }
    erg 
  end
=begin  ##
  # Speichert die berganstabelle in eine Datei, dessen Pfad der Parameter ist.
  # Sollte die Datei nicht vorhanden sein, wird sie erstellt, sonst berschrieben
  def save(file)
	 require 'yaml'
	 File.open(file,"w"){|data|
	   YAML.dump(self,data)
	 }
  end
  
  ##
  # Es wird eine bergangstabelle aus dem bergebenem Pfad geladen und zurckgeliefert.
  def load(file)
	require 'yaml'
	tmp = YAML::load(File.open(file))
	@namehs = tmp.getnamehs
	@tabelle = tmp.gettabelle
	@uSymbole = tmp.getsymbole
	@zusths = tmp.zusths
  end
=end
  
end


##
# Diese Klasse reprsentiert die Aktionstabelle eines (S)LR-Automaten.
# Sie erbt von ATable und arbeitet deswegen mit Automaten im DEA-Format.
# Der Aktionstabelle kann entnommen werden, abhngig vom aktuellen Zustand, 
# welche Aktion durch gefhrt werden kann.
# Denkbar wren:
# * shift : zum Lesen eines Zeichen
# * reduce[Item] : Reduktion mit der Regel Item
# * accept : Ist der "Endzustand", der eigentlich nur ein reduce[ S' -> S], wobei S der Startzustand der Grammatik ist
# * error : wenn kein bergang mglich ist.
# Dazu mehr im Buch: "Syntaxbassierte Programmierwerkzeuge" von Lothar Schmitz, Seite 178.
# Zustzlich werden noch die Konflikte errechnet, die auch abgerufen werden knnen.

class ATable < UTable

  ##
  # Der Konstruktor erwartet einen (S)LR-Automaten im DEA-Format.
  # Die Aktionstabelle wird intern in einem zweidimensionalem Hash berechnet.
  # 
  def initialize(dea)
    # Alle Zustnde in einem Hash speichern und ihnen einen "ordentlichen" Namen
    # nach folgendem Muster zuweissen: K[i] wobei i eine Nummer ist.
    # Eine Tablle ist ein Hash, mit Namen von Zustnden (das ist der Key), denen die Aktionen zugewiesen werden.
    # Eine Aktion kann sein (s.S. 155 u. 178):
    # * Reduktion
    # * Accept
    # * Shift
    # * Error
    super
		
    @conflictArray = Array.new
    conflictZust = Array.new
    # in der Tabelle alle Eintrge durchlaufen und alle NonTerminale Spalten entfernen
    @tabelle.each{|zKey,zeile|
     hasshift = false
     hasreduk = false
      @uSymbole.each_key{|symb|
        countReduk=0
        #puts symb.class
        if !symb.terminal
          # Wir haben ein NonTerminal
          # also aus der Tabelle entfernen und zusttzlich aus der Tabelle uSymb entfernen
          zeile.delete(symb)
          @uSymbole.delete(symb)
        else
          # Wenn es ein begang zu einem Zustand ist, dann wird es ein Shift
          if zeile[symb.to_s] != "-"
            zeile[symb.to_s] = "shift"
            hasshift = true
          else 
            # Berechnen, ob es ein Reduce oder ein Error ist
            # iterieren durch die Items eines Zustands
            @zusths[zKey].each{|item|
              # Is das vorletze Zeichen ein Punkt(:=46), dann ist es eine Reduktion
              if item.to_s[-2] == 46
                
                if item.getRechtskontext.enthaelt?(symb.to_s) or item.getRechtskontext.length == 0
                countReduk += 1
                hasreduk = true
                @tabelle[zKey][symb.to_s] = "red#{item.to_s}"
								if symb.to_s == "#" and zeile[symb.to_s].include? "red[S' ->"
									zeile[symb.to_s] = "accept"
								end
				end				
              end
            }
          end
        end
      #Alle Zustnde mit mglichen Konflikten werden in einem Array abgelegt.
      if (hasshift == true and hasreduk) or countReduk > 1
       (conflictZust.include?(@zusths[zKey])) ? () : (conflictZust.push(@zusths[zKey]))
      end
      }     
      zeile.default= "error"  
    }
    
    #Das Array mit den Konfliktzustnden genauer berprfen, ob es wirklich Konflikte sind.
    conflictZust.each{|zustand|
      #berprfen ob es ein Shift-Reduce-Konflikt ist
      zustand.each{|item|
        nextSymb = item.naechstesZeichen      
        zNeu = zustand.items.copy.array.delete_if{|i| i == item}
        #wenn nextSymb nicht nil ist, dann ist es ein Shift, sonst Reduce
        if nextSymb != nil
        # berprfen, ob in der Menge der Items des Zustands, das nchste Symbol als Rechtskontext vorkommt
        # wenn ja dann ist es ein Shift-Reduce Konflikt.
        zNeu.each{|item2|
          (item2.getRechtskontext.enthaelt?(nextSymb) and item2.to_s[-2] == 46) ? (zn=@zusths.invert[zustand];@conflictArray.push("#{zn} : Shift-Reduce-Konflikt")) : ()
        }
        else
        # berprfen, ob ein Reduce-Reduce-Konflikt auftritt
        # Das liegt vor, wenn ein Zeichen aus dem Rechtskontext des aktuellen Items
        # auch im Rechtskontext der anderen Items vorhanden ist und es mssen beide Reduktionen sein
        item.getRechtskontext.each{|symb|
          zNeu.each{|item2|
            (item2.getRechtskontext.enthaelt?(symb) and item2.to_s[-2] == 46) ? (zn=@zusths.invert[zustand];@conflictArray.push("#{zn} : Reduce-Reduce-Konflikt")) : ()
          }
        }
        end   
      }
      
    }    
  end
  
  ##
  # Liefert den Array mit den Konfliktmeldungen. Diese liegen in einem String vor.
  # Bsp. ["K1 : Shift-Reduce-Konflikt", "K2 : Reduce-Reduce-Konflikt"]
  def getConflicts
   @conflictArray
  end
  
  ##
  # Liefert einen boolschen Wert, ob der (S)LR-Automat konfliktfrei ist.
  def isConflictfree
   @conflictArray.empty?
  end
  
  ##
  # Liefert einen String, in dem alle Konflikte aufgelistet sind.
  # Sollte kein Konflikt vorliegen, steht dieses im String drin.
  def printConflicts
    if isConflictfree 
      return "Der Automat ist konfliktfrei."
    else
      confStr = String.new
      @conflictArray.each { |conflict|
        confStr += conflict +", "
      }
      return confStr[0 .. -3] 
     end
  end
end

=begin
  #c = Grammar.new("S -> b S d\n | b S c S d \n | a")
  c = Grammar.new("E -> E + T \n |T \n\n T -> T * id \n |id \n\n")
  a = LR_0.new(c)
  puts "\n"
  puts (u = UTable.new(a)).to_s
  puts "\n" + (tab = ATable.new a).to_s
  puts tab.printConflicts
  puts "-" *10
  b = SLR_1.new(c)
  puts "\n"
  puts (UTable.new b).to_s
  puts (tm = ATable.new(b)).to_s
  puts tm.printConflicts
=end